"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.cache = void 0;
const fs = require("fs");
const index_1 = require("../index");
var cache;
(function (cache) {
    let CacheErrorMessages;
    (function (CacheErrorMessages) {
        CacheErrorMessages["locked"] = "Object locked";
        CacheErrorMessages["notFound"] = "Object not found";
    })(CacheErrorMessages = cache.CacheErrorMessages || (cache.CacheErrorMessages = {}));
    cache.redisPrefix = index_1.config.general.cachePrefix;
    cache.lockPrefix = index_1.config.general.lockPrefix;
    cache.lockDuration = index_1.config.general.lockDuration;
    cache.scanCount = index_1.config.redis.scanCount || 1000;
    cache.lockRetry = 100;
    function client() {
        if (index_1.api.redis.clients && index_1.api.redis.clients.client) {
            return index_1.api.redis.clients.client;
        }
        else {
            throw new Error("redis not connected, cache cannot be used");
        }
    }
    cache.client = client;
    let lockNameOverride;
    function lockName() {
        if (lockNameOverride)
            return lockNameOverride;
        return index_1.id;
    }
    cache.lockName = lockName;
    function overrideLockName(name) {
        lockNameOverride = name;
    }
    cache.overrideLockName = overrideLockName;
    /**
     * A generic method to find all keys which match a pattern.
     * @pattern likely has a * at the end of the string.  Other arguments are for recursion and not required.
     */
    async function getKeys(pattern, count = cache.scanCount, keysAry = [], cursor = 0) {
        // return client().keys(redisPrefix + "*");
        const [newCursor, matches] = await client().scan(cursor, "MATCH", pattern, "COUNT", count);
        if (matches && matches.length > 0)
            keysAry = keysAry.concat(matches);
        if (newCursor === "0")
            return keysAry;
        return cache.getKeys(pattern, count, keysAry, parseInt(newCursor));
    }
    cache.getKeys = getKeys;
    /**
     * Returns all the keys in redis which are under this Actionhero namespace.  Potentially very slow.
     */
    async function keys(optionalScopePrefix = "") {
        // return client().keys(redisPrefix + "*");
        return getKeys(cache.redisPrefix + optionalScopePrefix + "*");
    }
    cache.keys = keys;
    /**
     * Returns all the locks in redis which are under this Actionhero namespace.  Potentially slow.
     */
    async function locks(optionalScopePrefix = "") {
        // return client().keys(lockPrefix + "*");
        return getKeys(cache.lockPrefix + optionalScopePrefix + "*");
    }
    cache.locks = locks;
    /**
     * Returns the number of keys in redis which are under this Actionhero namespace.  Potentially very slow.
     */
    async function size(pattern = cache.redisPrefix + "*") {
        let length = 0;
        const keys = await cache.getKeys(pattern);
        if (keys)
            length = keys.length;
        return length;
    }
    cache.size = size;
    /**
     * Removes all keys in redis which are under this Actionhero namespace.  Potentially very slow.
     * Returns the deleted keys.
     */
    async function clear(pattern = cache.redisPrefix + "*") {
        const keys = await cache.getKeys(pattern);
        const pipelineArgs = [];
        keys.forEach((key) => {
            pipelineArgs.push(["del", key]);
        });
        await client().pipeline(pipelineArgs).exec();
        return keys;
    }
    cache.clear = clear;
    /**
     * Write the current concents of redis (only the keys in Actionhero's namespace) to a file.
     */
    async function dumpWrite(file) {
        const data = {};
        const jobs = [];
        const keys = await cache.keys();
        keys.forEach((key) => {
            jobs.push(client()
                .get(key)
                .then((content) => {
                data[key] = content;
            }));
        });
        await Promise.all(jobs);
        fs.writeFileSync(file, JSON.stringify(data));
        return keys.length;
    }
    cache.dumpWrite = dumpWrite;
    /**
     * Load in contents for redis (and api.cache) saved to a file
     * Warning! Any existing keys in redis (under this Actionhero namespace) will be removed.
     */
    async function dumpRead(file) {
        const jobs = [];
        await cache.clear();
        const fileData = fs.readFileSync(file).toString();
        const data = JSON.parse(fileData);
        const count = Object.keys(data).length;
        const saveDumpedElement = async (key, content) => {
            await client().set(key, content, "KEEPTTL");
        };
        Object.keys(data).forEach((key) => {
            const content = data[key];
            jobs.push(saveDumpedElement(key, content));
        });
        await Promise.all(jobs);
        return count;
    }
    cache.dumpRead = dumpRead;
    /**
     * Load an item from the cache.  Will throw an error if the item named by `key` cannot be found.
     * Automatically handles `api.cache.redisPrefix`
     */
    async function load(key, options = {}) {
        let cacheObj;
        let lockOk = await cache.checkLock(key, options.retry);
        if (lockOk !== true)
            throw new Error(CacheErrorMessages.locked);
        let cachedStringifiedObjet = await client().get(`${cache.redisPrefix}${key}`);
        try {
            cacheObj = JSON.parse(cachedStringifiedObjet);
        }
        catch (e) { }
        if (!cacheObj)
            throw new Error(CacheErrorMessages.notFound);
        lockOk = await cache.checkLock(key, options.retry);
        if (lockOk !== true)
            throw new Error(CacheErrorMessages.locked);
        if (options.expireTimeMS) {
            await client().pexpire(cache.redisPrefix + key, options.expireTimeMS);
            return {
                key,
                value: cacheObj.value,
                createdAt: cacheObj.createdAt,
            };
        }
        else {
            return {
                key,
                value: cacheObj.value,
                createdAt: cacheObj.createdAt,
            };
        }
    }
    cache.load = load;
    /**
     * Delete an item in the cache.  Will throw an error if the item named by `key` is locked.
     * Automatically handles `api.cache.redisPrefix`
     */
    async function destroy(key) {
        const lockOk = await cache.checkLock(key, null);
        if (!lockOk)
            throw new Error(CacheErrorMessages.locked);
        const count = await client().del(cache.redisPrefix + key);
        let response = true;
        if (count !== 1) {
            response = false;
        }
        return response;
    }
    cache.destroy = destroy;
    /**
     * Save an item in the cache.  If an item is already in the cache with the same key, it will be overwritten.  Throws an error if the object is already in the cache and is locked.
     * Automatically handles `api.cache.redisPrefix`
     */
    async function save(key, value, expireTimeMS) {
        const cacheObj = {
            value: value,
            createdAt: new Date().getTime(),
        };
        const lockOk = await cache.checkLock(key, null);
        if (!lockOk)
            throw new Error(CacheErrorMessages.locked);
        await client().set(cache.redisPrefix + key, JSON.stringify(cacheObj));
        if (expireTimeMS) {
            await client().pexpire(cache.redisPrefix + key, expireTimeMS);
        }
        return true;
    }
    cache.save = save;
    /**
     * Push an item to a shared queue/list in redis.
     * Automatically handles `api.cache.redisPrefix`
     */
    async function push(key, item) {
        const object = JSON.stringify({ data: item });
        await client().rpush(cache.redisPrefix + key, object);
        return true;
    }
    cache.push = push;
    /**
     * Pop (get) an item to a shared queue/list in redis.
     * Automatically handles `api.cache.redisPrefix`
     */
    async function pop(key) {
        const object = await client().lpop(cache.redisPrefix + key);
        if (!object) {
            return null;
        }
        const item = JSON.parse(object);
        return item.data;
    }
    cache.pop = pop;
    /**
     * Check how many items are stored in a shared queue/list in redis.
     */
    async function listLength(key) {
        return client().llen(cache.redisPrefix + key);
    }
    cache.listLength = listLength;
    /**
     * Lock an item in redis (can be a list or a saved item) to this Actionhero process.
     */
    async function lock(key, expireTimeMS = cache.lockDuration) {
        const lockOk = await cache.checkLock(key, null);
        if (!lockOk) {
            return false;
        }
        const result = await client().setnx(cache.lockPrefix + key, lockName());
        if (!result) {
            return false;
        } // value was already set, so we cannot obtain the lock
        await client().expire(cache.lockPrefix + key, Math.ceil(expireTimeMS / 1000));
        return true;
    }
    cache.lock = lock;
    /**
     * Unlock an item in redis (can be a list or a saved item) which was previously locked by this Actionhero process.
     */
    async function unlock(key) {
        const lockOk = await cache.checkLock(key, null);
        if (!lockOk) {
            return false;
        }
        await client().del(cache.lockPrefix + key);
        return true;
    }
    cache.unlock = unlock;
    async function checkLock(key, retry = false, startTime = new Date().getTime()) {
        const lockedBy = await client().get(cache.lockPrefix + key);
        if (lockedBy === lockName() || lockedBy === null) {
            return true;
        }
        else {
            const delta = new Date().getTime() - startTime;
            if (!retry)
                return false;
            else if (typeof retry === "boolean")
                return retry;
            else if (delta > retry)
                return false;
            await index_1.utils.sleep(cache.lockRetry);
            return cache.checkLock(key, retry, startTime);
        }
    }
    cache.checkLock = checkLock;
})(cache || (exports.cache = cache = {}));
